
/* Don't set the platform font family
   - We want it to match the system font on different Linux DEs
   - QT should set it correctly on Windows/MacOS
   - We need to construct the default font in .cpp, so it won't match the CSS

   The default platform font might not have a "thin" or "light" style, these
   are the ones I know about:

   Windows 10/11: Segoe UI      thin:yes light:yes
   Ubuntu:        Ubuntu Sans   thin:yes light:yes
   KDE:           Noto Sans     thin:no  light:no
   KDE:           DejaVu Sans   thin:no  light:no
   GNOME:         Cantarell     thin:yes light:yes
   MacOS:         San Francisco thin:yes light:yes
*/

QWidget {
    /* Note this won't always match the platform default font however we must do it
       because other font sizes used here and in .cpp are based on it */
    font-size: 11pt;
}

QToolTip {
    font-size: 10pt;
}

/* batch */
nmc--DkBatchInput QTextEdit {
	border: none;
}

QLabel#batchContentTitle {
	text-transform: uppercase;
    font-style: thin;
    font-size: 24pt; /* match DkPreferenceTitle */
	font-weight: 100;
}

QLabel#batchContentInfo {
    padding-bottom: 20px;
    font-style: italic;
    font-size: 10pt; /* should match DkBatchTabButton::paintEvent() */
}

QLabel#BatchInfo {
    font-style: italic;
    color: #fff;
}

QLabel#FileNamePreviewLabel {
    font-size: 11pt;
}

nmc--DkPreferenceWidget nmc--DkResizableScrollArea,
nmc--DkBatchWidget nmc--DkResizableScrollArea {
	border: none;
}

nmc--DkBatchButtonsWidget QPushButton:hover,
nmc--DkBatchButtonsWidget QPushButton:checked:hover {
	background-color: rgba(255, 255, 255, 30);
} 

nmc--DkBatchButtonsWidget QPushButton:pressed,
nmc--DkBatchButtonsWidget QPushButton:checked:pressed {
	background-color: HIGHLIGHT_LIGHT;
}

nmc--DkBatchButtonsWidget QPushButton:checked {
	background-color: none;
}

nmc--DkBatchButtonsWidget QPushButton {
	border: none;
}

QPushButton#flatWhite {
	border: none;
	padding: 0;
	margin: 0;
}

QPushButton#flatWhite:hover {
	background-color: rgba(255,255,255,.2);
}

QPushButton#flatWhite:pressed {
	background-color: rgba(255,255,255,.4);
}

/* nomacs-if-colorscheme-eq-dark */
QTabBar::close-button {
  image: url(:/nomacs/img/close-white.svg);
}
QTabBar::close-button:hover {
  image: url(:/nomacs/img/close-white-hover.svg);
}
/* nomacs-else */
QTabBar::close-button {
  image: url(:/nomacs/img/close.svg);
}
QTabBar::close-button:hover {
  image: url(:/nomacs/img/close-hover.svg);
}
/* nomacs-endif */

/* prevent transparent background in central widgets,
   dock widgets/status bar is handled by paintEvent() */
nmc--DkNoMacsFrameless QTabBar,
nmc--DkNoMacsFrameless QToolBar,
nmc--DkNoMacsFrameless nmc--DkBatchWidget,
nmc--DkNoMacsFrameless nmc--DkPreferenceWidget,
nmc--DkNoMacsFrameless nmc--DkRecentFilesWidget QScrollArea > QWidget > QWidget {
  background-color: BACKGROUND_COLOR;
}

QPushButton#DkPlayerButton {
	margin: 0px;
  border: none;
	background-color: HUD_BACKGROUND_COLOR;
}

QPushButton#hudNavigationButton:hover,
QPushButton#DkPlayerButton:hover{
	background-color: rgba(0,0,0,.6);
} 

QPushButton#hudNavigationButton:pressed,
QPushButton#DkPlayerButton:pressed {
	background-color: HIGHLIGHT_LIGHT;
}

QPushButton#hudNavigationButton {
    border: none;

    min-height: 32px;
    max-height: 128px;
}

QPushButton#DkRestartButton {
	border: none;
    padding: 8px;
}

QPushButton#DkRestartButton:hover{
    background-color: rgba(255,255,255,.2);
}

QPushButton#DkRestartButton:pressed {
	background-color: HIGHLIGHT_LIGHT;
}

QToolBar {
	border-top: none;
	border-bottom: 1px solid #cdcdcd;
}

QToolBar#cropToolBar {
	spacing: 3px; 
	padding: 3px;
}

QStatusBar#statusBarWithGradient {
	border-top: none; 
	background: QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #edeff9, stop: 1 #bebfc7); 
}
  
QLabel#thumbsTitle {
	background: rgba(0,0,0,30);
}

QGraphicsView#DkThumbsView {
	border: none;
}

/* FIXME: the stylesheet is not reevaluted when entering/leaving fullscreen,
          so it keeps the background it started with */
QMainWindow[fullScreen="false"] QGraphicsView#DkThumbsView {
    background-color: BACKGROUND_COLOR;
}

QMainWindow[fullScreen="true"] QGraphicsView#DkThumbsView {
    background-color: SLIDESHOW_COLOR;
}

nmc--DkNoMacsFrameless[fullScreen="false"] QGraphicsView#DkThumbsView {
    background-color: transparent;
}

nmc--DkNoMacsFrameless[fullScreen="true"] QGraphicsView#DkThumbsView {
    background-color: HUD_BACKGROUND_COLOR;
}

nmc--DkNoMacsFrameless nmc--DkThumbsView > QWidget { /* scrollbar background */
    background-color: BACKGROUND_COLOR;
}

QToolButton#DkThumbToolButton::menu-indicator { 
	image: none; 
}

QLabel#statusBarLabel {
	color: #555;
}

/* 
QToolButton:hover{
	border: none; 
	background-color: rgba(255,255,255,90);
} 

QToolButton:pressed{
	margin: 0px; 
	border: none; 
	background-color: HIGHLIGHT_COLOR;
}
 */
 
/* ViewPort */
QGraphicsView#DkBaseViewPortGradient {
	border-style: none; 
	background: QLinearGradient(x1: 0, y1: 0.7, x2: 0, y2: 1, stop: 0 #edeff9, stop: 1 #d9dbe4);
 }
 
QGraphicsView#DkBaseViewPortDefaultColor {
	border-style: none; 
    background-color: palette(window);  /* why can't this be BACKGROUND_COLOR */
}

QGraphicsView#DkBaseViewPort {
	border-style: none; 
	background-color: BACKGROUND_COLOR;
} 

/* Batch Widgets */
QLabel#DkBatchTitle {
	font-size: 16px;
}

/* Dialogs */
QLabel#DkSplashInfoLabel {
    font-size: 12px;
}

QPushButton#DkSplashCloseButton {
    image: url(":/nomacs/img/close.svg");
    padding: 4px;
    margin: 8px;
    max-width: 1em;
    max-height: 1em;
}

QPushButton#DkSplashCloseButton:hover {
    image: url(":/nomacs/img/close-hover.svg");
}

QPushButton#DkSplashCopyInfoButton {
  image: url(":nomacs/img/copy.svg");
  padding: 2px;
  margin: 0px;
  max-width: 1em;
  max-height: 1em;
}

QPushButton#DkSplashCopyInfoButton:hover {
  background-color: #ddd;
}

QLineEdit#DkWarningEdit {
	color: #000;
}

QLabel#DkDecentInfo[warning="true"],
QLineEdit#DkWarningEdit[error="true"] {
	color: #9f1d1d;
}

QListView#resultListView[empty="true"] {
	color: #777777; 
	font-style: italic;
}

QListWidget#historyList {
	font-size: 12pt;
}

QListWidget#historyList::item:disabled {
	font-style: italic;
}

/* HUD Widgets */
QSlider#slZoom::groove:horizontal {
    height: 5px;
    border: 1px solid HUD_FOREGROUND_COLOR;
}

QSlider#slZoom::handle:horizontal {
    width: 10px;
    border-radius: 3px;
	background: HUD_FOREGROUND_COLOR;
}

QLabel#DkOverview {
	background: HUD_BACKGROUND_COLOR;
}

QLabel#DkOverviewSliderWidget {
	background: HUD_BACKGROUND_COLOR;
}

QLabel#commentTitleLabel {
	color: HUD_FOREGROUND_COLOR;
	margin: 0; 
	padding: 0; 
    font-size: 18pt;
	font-weight: light;
}

nmc--DkCommentWidget {
	background-color: HUD_BACKGROUND_COLOR;
    min-width: 15em; /* originally 220x150 px */
    max-width: 15em;
    min-height: 8em;
    max-height: 8em;
}

nmc--DkCommentWidget > nmc--DkCommentTextEdit {
	border: 0; 
	padding-top: 10px; 
	background-color: rgba(0,0,0,0);
	border-top: 2px dotted HUD_FOREGROUND_COLOR; 
	color: HUD_FOREGROUND_COLOR; 
    font-size: 12pt;
	selection-background-color: HIGHLIGHT_COLOR; 
}

QLabel#DkMetaDataHUDTitle {
	color: HUD_FOREGROUND_COLOR; 
	margin: 0; 
	font-size: 21px; 
	font-weight: light;
}

QLabel#DkSeparator {
	border-top: 2px dotted HUD_FOREGROUND_COLOR; 
}

nmc--DkFolderScrollBar {
	background-color: HUD_BACKGROUND_COLOR;
	max-height: 15px;
	border: 1px solid #000;
}

nmc--DkFolderScrollBar:handle {
	background-color: HIGHLIGHT_LIGHT;
	width: 30px;
}

nmc--DkFolderScrollBar:handle:hover {
	background-color: HIGHLIGHT_COLOR;
}

nmc--DkFolderScrollBar::groove:horizontal {
	background-color: rgba(0,0,0,0);
}

nmc--DkFileInfoLabel {
	background-color: HUD_BACKGROUND_COLOR;
	color: HUD_FOREGROUND_COLOR;
}

QLabel#DkMetaDataLabel {
	color: HUD_FOREGROUND_COLOR;
}

QLabel#DkMetaDataKeyLabel {
	color: HUD_FOREGROUND_COLOR;
}

nmc--DkManipulatorWidget > QScrollArea,
nmc--DkMetaDataHUD > QScrollArea {
	background-color: transparent;
	border: none;
}

nmc--DkManipulatorWidget > QScrollArea > QWidget > QWidget,
nmc--DkMetaDataHUD > QScrollArea > QWidget > QWidget {
	background-color: transparent;
	border: none;
}

nmc--DkMetaDataHUD {
	background-color: HUD_BACKGROUND_COLOR;
}

QPushButton#removeFileButton {
	margin: 0 0 0 0; 
	padding: 0 0 0 0; 
	border: none;
}

QPushButton#removeFileButton:hover {
	background-color: rgba(255,255,255,.4);
}
 
/* preferences, adjustments, batch dark panel */
QScrollArea#DkPreferenceTabs,
QWidget#DkPreferenceTabs,
QWidget#DkBatchTabs {
/* nomacs-if-colorscheme-eq-dark */
  background-color: nomacsBlend(BACKGROUND_COLOR,0,0,0,50);
/* nomacs-else */
  background-color: nomacsBlend(BACKGROUND_COLOR,0,0,0,128);
/* nomacs-endif */
}

/* fixes padding view issue of tab entries */
QScrollArea#DkPreferenceTabsScroller {
    min-width: 270px;
}

QPushButton#DkTabEntryWidget,
nmc--DkBatchTabButton {
	color: #fff;
  font-style: thin;
	font-size: 14pt;
	padding: 10px 20px;
	text-align: left;
	border: none;
}

nmc--DkBatchTabButton {
	min-width: 250px;
	padding-bottom: 30px;
}

QPushButton#DkTabEntryWidget:hover,
QPushButton#DkTabEntryWidget:checked:hover,
nmc--DkBatchTabButton:hover,
nmc--DkBatchTabButton:checked:hover {
	background-color: rgba(255, 255, 255, 30);
}

QPushButton#DkTabEntryWidget:checked,
nmc--DkBatchTabButton:checked {
	background-color: rgba(0, 0, 0, 80);
	border: none;	/* clear border styles first */
	border-left: 4px solid HIGHLIGHT_COLOR;
}

QLabel#DkPreferenceTitle {
	text-transform: uppercase;
	padding: 10px 20px;
    font-style: thin;
    font-size: 24pt; /* match batchContentTitle */
	font-weight: 100;
}

QPushButton#infoButton {
    font-style: thin;
	font-size: 16pt;
	font-weight: 100;
	text-align: left;
	padding-left: 20px;
	border: none;
}

QPushButton#infoButton:hover {
	color: HIGHLIGHT_COLOR;
}

QPushButton#infoButton:pressed {
	color: HIGHLIGHT_LIGHT;
}

QWidget#DkPreferenceWidget {
	margin-top: 10px; /* not working?! */
	margin-left: 10px;
}

QLabel#subTitle {
    font-style: thin;
	margin-top: 10px;
	font-size: 14pt;
}

/* preferences widgets */
QComboBox#languageCombo,
QSpinBox#numThreads,
QPushButton#logFolder,
nmc--DkDisplayPreference QSpinBox,
nmc--DkDisplayPreference QDoubleSpinBox,
nmc--DkDisplayPreference QComboBox {
	max-width: 200px;
}

QLabel#settingsInfoLabel {
	padding-top: 6px;
    font-style: italic;
}

/*
nmc--DkGroupWidget QCheckBox,
nmc--DkGroupWidget QRadioButton,
nmc--DkGroupWidget QPushButton,
nmc--DkGroupWidget QSpinBox,
nmc--DkGroupWidget QDoubleSpinBox,
nmc--DkGroupWidget QComboBox,
nmc--DkGroupWidget QLineEdit,
nmc--DkGroupWidget QLabel {
    font-size: 11pt;
}
*/

nmc--DkProgressBar {
	max-height: 3px;
}

/* plugins */
nmc--DkDescriptionEdit {
	border: none;
	background-color: rgba(0,0,0,0);
}

/* edit image */
nmc--DkBaseManipulatorWidget#darkManipulator QWidget {
/* nomacs-if-colorscheme-eq-dark */
  background-color: nomacsBlend(BACKGROUND_COLOR,0,0,0,100);
/* nomacs-else */
  background-color: nomacsBlend(BACKGROUND_COLOR,0,0,0,128);
/* nomacs-endif */
}

nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox,
nmc--DkBaseManipulatorWidget#darkManipulator QLabel,
nmc--DkBaseManipulatorWidget#darkManipulator QCheckBox,
nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox,
nmc--DkBaseManipulatorWidget#darkManipulator QComboBox,
nmc--DkBaseManipulatorWidget#darkManipulator QLineEdit {
	color: #fff;
}

nmc--DkBaseManipulatorWidget#darkManipulator QLineEdit,
nmc--DkBaseManipulatorWidget#darkManipulator QComboBox,
nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox,
nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox {
  border: none;
	border-bottom: 1px solid #fff;
}

nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox::up-button,
nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox::down-button,
nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox::up-button,
nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox::down-button {
	border: none;
  background-color: transparent;
  width: 0px;
  height: 0px;
}	

nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox::up-arrow,
nmc--DkBaseManipulatorWidget#darkManipulator QDoubleSpinBox::down-arrow,
nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox::up-arrow,
nmc--DkBaseManipulatorWidget#darkManipulator QSpinBox::down-arrow {
  image: none;
}

nmc--DkBaseManipulatorWidget#darkManipulator QSlider::groove:horizontal {
	border: none; 
	height: 3px; 
	margin: 2px 0;
	background: #fff;
}

nmc--DkBaseManipulatorWidget#darkManipulator QSlider::handle:horizontal {
	background-color: #fff; 
	width: 6px; 
	margin:-8px 0px -8px 0px;
	border: 2px solid #333;
}

nmc--DkBaseManipulatorWidget#darkManipulator {
  /* background-color: palette(light); */
  /* border-radius: 7px; */
}

/* DkManipulationWidget */
nmc--DkBaseManipulatorWidget#darkManipulator QSlider#DkBrightnessSlider::groove:horizontal {
	background: qlineargradient(x1:0, y1:0, x2:1, y2:0, stop:0 #3c3c3c, stop:1 #c8c8c8);
}

nmc--DkBaseManipulatorWidget#darkManipulator QSlider#DkSaturationSlider::groove:horizontal {
	background: qlineargradient(x1:0, y1:0, x2:1, y2:0, stop:0 #ffffff, stop:1 #00ffff);
}

nmc--DkBaseManipulatorWidget#darkManipulator QSlider#DkHueSlider::groove:horizontal {
	background: qlineargradient(x1:0, y1:0, x2:1, y2:0, stop:0 #ff0000, stop:0.167 #ffff00, stop:0.333 #00ff00, stop:0.5 #00ffff, stop:0.666 #0000ff, stop:0.833 #ff00ff, stop:1 #ff0000);
}

nmc--DkBaseManipulatorWidget#darkManipulator QComboBox {
    font-size: 12pt;
	min-height: 30px;
}

nmc--DkBaseManipulatorWidget#darkManipulator QComboBox::drop-down {
	image: url(:/nomacs/img/down-white.svg);
	border: none;
	width: 16px;
	height: 16px;
	padding: 5px;
}

nmc--DkBaseManipulatorWidget#darkManipulator QComboBox::drop-down:hover {
	background-color: rgba(255, 255, 255, 30);
}

nmc--DkBaseManipulatorWidget#darkManipulator QComboBox QListView {
	border: 1px solid #444;
    font-size: 12pt;
}

nmc--DkBaseManipulatorWidget#darkManipulator QComboBox QListView::item {
	color: #fff;
	padding: 5px;
  selection-color: #fff;
  selection-background-color: HIGHLIGHT_COLOR;
}

QLabel#DkManipulatorSettingsTitle {
    font-style: thin;
	font-size: 14pt;
	color: #fff;
	margin-bottom: 5pt;
}

QSlider#cpHueSlider::groove:vertical {
	background: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #ff0000, stop:0.167 #ff00ff, stop:0.333 #0000ff, stop:0.5 #00ffff, stop:0.666 #00ff00, stop:0.833 #ffff00, stop:1 #ff0000);
}

QSlider#cpHueSlider::handle:vertical {
	border: 0px solid #333;
	border-top: 2px solid #333;
	border-bottom: 2px solid #333;
	height: 5px;
}

QLabel#summaryMeta {
	margin-left: 20px;
	qproperty-alignment: AlignRight;
}

QListWidget#profileList {
	font-size: 14pt;
	max-width: 400px;
}

/*nmc--DkMetaDataSelection QScrollArea QWidget {
	background-color: #fff;
}*/

QLabel#DkLabelBg {
	color: HUD_FOREGROUND_COLOR;
	background-color: HUD_BACKGROUND_COLOR;
	padding: 5px;
}

/* DkRecentFiles */
nmc--DkRecentFilesWidget QScrollArea {
	border: none;
	background-color: transparent;
}

nmc--DkRecentDirWidget {
	background-color: rgba(255, 255, 255, 100);
	border: 1px solid rgba(255, 255, 255, 150);
	border-radius: 7px;
}

nmc--DkRecentDirWidget::hover {
	background-color: rgba(255, 255, 255, 150);
}

QLabel#recentFilesTitle {
    font-style: thin;
	font-size: 16pt;
	font-weight: 100;
}

QLabel#recentFilesPath {
	font-style: italic;
}

QPushButton#load_dir,
QPushButton#pin,
QPushButton#remove {
	border: none;
	padding: 4px;
}

QPushButton#load_dir:hover,
QPushButton#pin:hover,
QPushButton#remove:hover {
	background-color: HIGHLIGHT_LIGHT;
} 

QPushButton#load_dir:pressed,
QPushButton#pin:pressed,
QPushButton#remove:pressed {
	background-color: HIGHLIGHT_LIGHT;
}

nmc--DkThumbPreviewLabel {
	padding: 1px;
}

nmc--DkThumbPreviewLabel[empty="true"] {
	background-color: #ddd;
}

nmc--DkThumbPreviewLabel:hover {
	border: 1px solid HIGHLIGHT_COLOR;
}

QDoubleSpinBox#sbZoom {
	margin: 0px;
	padding: 0px; 
	color: HUD_FOREGROUND_COLOR;
	background-color: rgba(0,0,0,0); 
	border: none;
	selection-background-color: HIGHLIGHT_COLOR;
}

QPushButton#displayButton {
    font-style: light;
	font-size: 16pt;
	font-weight: 100;
}

QPushButton#displayButton:hover {
	background-color: HIGHLIGHT_LIGHT;
}

QPushButton#displayButton:checked {
	border: 1px solid HIGHLIGHT_COLOR;
}

/* picks the wrong color on system theme; doesn't seem to be needed anymore
QToolBar {
    background-color: BACKGROUND_COLOR;
}
*/
